__all__ = [
    "EncodeError", "ParseError",
    "Asn1Data", "Class", "Structure", "Identifier",
    "decode", "encode",
]

from enum import IntEnum

from snmp.exception import *
from snmp.typing import *
from snmp.utils import subbytes

Asn1Data = Union[bytes, subbytes]

@final
class EncodeError(SNMPException):
    """Failure to encode a payload under ASN.1 Basic Encoding Rules."""
    pass

@final
class ParseError(IncomingMessageError):
    """Failure to translate a byte string into an ASN.1 object."""
    pass

@final
class Class(IntEnum):
    """Named constants for the class bits of an ASN.1 BER identifier."""
    UNIVERSAL         = 0
    APPLICATION       = 1
    CONTEXT_SPECIFIC  = 2
    PRIVATE           = 3

@final
class Structure(IntEnum):
    """Named constants for the constructed bit of an ASN.1 BER identifier."""
    PRIMITIVE     = 0
    CONSTRUCTED   = 1

TIdentifier = TypeVar("TIdentifier", bound="Identifier")

@final
class Identifier(NamedTuple):
    """Represents an ASN.1 BER identifier."""
    cls: Class
    structure: Structure
    tag: int

    @classmethod
    def decode(cls: Type[TIdentifier], data: subbytes) -> TIdentifier:
        """Extract the identifier from an ASN.1 BER string.

        This function decodes the identifier portion of a BER string an returns
        it as an :class:`Identifier` object. As a side-effect, it advances the
        start of the `data` argument to point to the byte immediately after the
        identifier.
        """
        try:
            byte = data.consume()
        except IndexError as err:
            raise ParseError("Missing identifier") from err

        class_      = (byte & 0xc0) >> 6
        structure   = (byte & 0x20) >> 5
        tag         = (byte & 0x1f)

        if tag == 0x1f:
            tag = 0
            byte = 0x80
            while byte & 0x80:
                try:
                    byte = data.consume()
                except IndexError as err:
                    raise ParseError("Incomplete identifier") from err

                tag <<= 7
                tag |= byte & 0x7f

        return cls(Class(class_), Structure(structure), tag)

    def encode(self) -> bytes:
        """Encode an Identifier under ASN.1 Basic Encoding Rules."""
        byte = (
            ((self.cls       << 6) & 0xc0) |
            ((self.structure << 5) & 0x20)
        )

        arr = bytearray()
        if self.tag < 0x1f:
            byte |= self.tag
        else:
            byte |= 0x1f

            tag = self.tag
            indicator = 0
            while tag:
                arr.append(indicator | (tag & 0x7f))
                indicator = 0x80
                tag >>= 7

        arr.append(byte)
        return bytes(reversed(arr))

def decode_length(data: subbytes) -> int:
    """Decode the length field of an ASN.1 BER string.

    The provided `data` argument should contain most of a BER string,
    starting after the identifier. The most natural way to use it is to pass
    the same object first to :func:`Identifier.decode` and then to
    :func:`decode_length`. This function will decode the length field and
    return it as an :class:`int`, modifying the `data` argument to start
    immediately after the length field.
    """
    try:
        length = data.consume()
    except IndexError as err:
        raise ParseError("Missing length") from err

    if length & 0x80:
        n = length & 0x7f

        length = 0
        for i in range(n):
            try:
                byte = data.consume()
            except IndexError as err:
                raise ParseError("Incomplete length") from err

            length <<= 8
            length |= byte

    return length

def encode_length(length: int) -> bytes:
    """Encode the length of a message under ASN.1 Basic Encoding Rules."""
    if length < 0x80:
        return bytes([length])

    arr = bytearray()
    while length:
        arr.append(length & 0xff)
        length >>= 8

    if len(arr) > 0x7f:
        raise EncodeError("Length too large for definite-length encoding")

    arr.append(0x80 | len(arr))
    return bytes(reversed(arr))

@overload
def decode(
    data: Asn1Data,
    expected: None,
    leftovers: Literal[False],
    copy: Literal[False],
) -> Tuple[Identifier, subbytes]:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: None = None,
    leftovers: Literal[False] = False,
    copy: Literal[True] = True,
) -> Tuple[Identifier, bytes]:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: None,
    leftovers: Literal[True],
    copy: Literal[False],
) -> Tuple[Identifier, subbytes, subbytes]:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: None,
    leftovers: Literal[True],
    copy: Literal[True] = True,
) -> Tuple[Identifier, bytes, subbytes]:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: Identifier,
    leftovers: Literal[False],
    copy: Literal[False],
) -> subbytes:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: Identifier,
    leftovers: Literal[False] = False,
    copy: Literal[True] = True,
) -> bytes:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: Identifier,
    leftovers: Literal[True],
    copy: Literal[False],
) -> Tuple[subbytes, subbytes]:
    ...

@overload
def decode(
    data: Asn1Data,
    expected: Identifier,
    leftovers: Literal[True],
    copy: Literal[True] = True,
) -> Tuple[bytes, subbytes]:
    ...

def decode( # type: ignore[no-untyped-def]
    data: Asn1Data,
    expected: Optional[Identifier] = None,
    leftovers: bool = False,
    copy: bool = True,
):
    """Extract the contents of an ASN.1 BER string.

    This function has several options for what it can return depending on
    the values of its arguments. In the default case, given only the `data`
    argument, it will return a tuple, where the first element is the
    :class:`Identifier`, and the second contains the body of the message as
    a `bytes` object. In this case, the function expects to consume the
    entire message, and will raise a :class:`ParseError` if it contains more
    bytes than what the length field indicates. A value of ``True`` for the
    `leftovers` argument will change this behavior, so that it will return
    the leftover data in a :class:`snmp.utils.subbytes` object as an
    additional element of the returned tuple.

    The `expected` argument allows the caller to provide an
    :class:`Identifier`, telling the function what data type is expected. In
    this case, the body will be returned without the identifier, and the
    return value will not use a tuple (unless `leftovers` was given as
    ``True``). If the decoded object does not match the expected type, the
    function will raise a :class:`ParseError`.

    The `copy` argument allows the caller to specify whether the body of the
    message should be copied into its own :class:`bytes` object (the
    default), or use a :class:`snmp.utils.subbytes` object, in order to
    preserve a reference to the complete message.
    """
    data = subbytes(data)
    identifier = Identifier.decode(data)

    if expected is not None and identifier != expected:
        raise ParseError("Identifier does not match expected type")

    length = decode_length(data)
    tail = data.prune(length)

    if len(data) < length:
        raise ParseError("Incomplete value")
    elif not leftovers and tail:
        raise ParseError("Trailing bytes")

    body = data[:] if copy else data

    if expected is None:
        if leftovers:
            return identifier, body, tail
        else:
            return identifier, body
    else:
        if leftovers:
            return body, tail
        else:
            return body

def encode(identifier: Identifier, data: bytes) -> bytes:
    """Encode a message under ASN.1 Basic Encoding Rules."""
    return identifier.encode() + encode_length(len(data)) + data



__all__ = [
    "Integer32", "IpAddress", "Counter32", "Gauge32",
    "Unsigned32", "TimeTicks", "Opaque", "Counter64",
    "zeroDotZero",
]

from socket import inet_aton, inet_ntoa

from snmp.ber import *
from snmp.types import *
from snmp.typing import *
from snmp.utils import *

class Unsigned(Integer):
    SIGNED = False

@final
class Integer32(Integer):
    pass

@final
class IpAddress(OctetString):
    TYPE = Identifier(Class.APPLICATION, Structure.PRIMITIVE, 0)
    MIN_SIZE = 4
    MAX_SIZE = 4

    def __init__(self, addr: str) -> None:
        self.addr = addr

    def __repr__(self) -> str:
        return f"{typename(self)}({repr(self.addr)})"

    def equals(self, other: OctetString) -> bool:
        return self.data == other.data

    @property
    def data(self) -> bytes:
        try:
            return inet_aton(self.addr)
        except OSError as err:
            raise ValueError(f"Invalid IPv4 address: \"{self.addr}\"") from err

    @classmethod
    def interpret(cls, data: Asn1Data) -> "IpAddress":
        addr = data[:] if isinstance(data, subbytes) else data
        return cls(inet_ntoa(addr))

@final
class Counter32(Unsigned):
    TYPE = Identifier(Class.APPLICATION, Structure.PRIMITIVE, 1)

class Unsigned32(Unsigned):
    TYPE = Identifier(Class.APPLICATION, Structure.PRIMITIVE, 2)

@final
class Gauge32(Unsigned32):
    pass

@final
class TimeTicks(Unsigned):
    TYPE = Identifier(Class.APPLICATION, Structure.PRIMITIVE, 3)

@final
class Opaque(OctetString):
    TYPE = Identifier(Class.APPLICATION, Structure.PRIMITIVE, 4)

@final
class Counter64(Unsigned):
    BITS = 64
    TYPE = Identifier(Class.APPLICATION, Structure.PRIMITIVE, 6)

zeroDotZero = OID(0, 0)



__all__ = ["ComparableWeakRef", "NumberGenerator", "subbytes", "typename"]

from abc import ABCMeta, abstractmethod
from random import randint
import weakref

from snmp.typing import *

T = TypeVar("T")
V = TypeVar("V")    # V must support less-than comparison

@final
class ComparableWeakRef(Generic[T, V]):
    """Allow weakly-referenced objects to be used in sorted data structures.

    This class extends the behavior of the standard library's
    :class:`weakref` type by adding support for the less-than operator
    (:meth:`__lt__`). Upon construction, it will call the provided `key`
    function, with `obj` as the only argument, and the return value will be
    stored as the "comparison key". So long as the object remains alive, it
    will continue to call `key` with every comparison, and update the stored
    value. If the reference becomes invalid (i.e. the object is garbage-
    collected), then it will continue to use the last stored value for all
    future comparisons.
    """

    def __init__(self, obj: T, key: Callable[[T], V]) -> None:
        # obj could be garbage-collected as soon as this call returns, so it's
        # important to retrieve the value now, rather than initialize to None
        self._value = key(obj)

        self.key = key
        self.ref = weakref.ref(obj)

    @property
    def value(self) -> V:
        obj = self.ref()

        if obj is not None:
            self._value = self.key(obj)

        return self._value

    def __call__(self) -> Optional[T]:
        """Retrieve a reference to the wrapped object.

        If the referenced object is still alive, then it will be returned.
        Otherwise, this call will return ``None``.
        """
        return self.ref()

    def __lt__(self, other: "ComparableWeakRef[Any, V]") -> bool:
        """Compare this object to another ComparableWeakRef."""
        return cast(bool, self.value < other.value) # type: ignore[operator]

@final
class NumberGenerator:
    """Generate integers spanning a specific range.

    Given an integer size of `n` bits, an instance of this class will
    generate a sequence of length ``2^n``, where each `n`-bit integer
    appears exactly once. The first ``2^n-1`` numbers in the sequence are
    guaranteed to be non-zero; or, in other words, the last number in the
    sequence is always ``0``. After ``2^n`` iterations, the sequence repeats
    from the beginning.

    The `signed` argument determines whether the generated numbers should
    use two's-complement encoding (i.e. cover the range ``-(2^(n-1))`` to
    ``2^(n-1)-1``, inclusive), or unsigned encoding (``0`` to ``2^n-1``).
    """

    def __init__(self, n: int, signed: bool = True) -> None:
        half = 1 << (n-1)

        self.previous = 0
        self.range = half << 1
        self.step = 2 * randint(1, half) - 1
        self.wrap = self.range - 1

        if signed:
            self.wrap -= half

    def __iter__(self) -> Iterator[int]:
        return self

    def __next__(self) -> int:
        """Give the next number in the sequence."""
        self.previous += self.step

        if self.previous > self.wrap:
            self.previous -= self.range

        return self.previous

@final
class subbytes:
    """Operate on a slice of a bytes-like object without copying any data.

    This class represents a sequence of bytes, similarly to how data is
    represented in a bytes-like object. Unlike a bytes-like object, however,
    a :class:`subbytes` object does not store this data in a memory block of
    its own. It holds a reference to a real bytes-like object, and acts on a
    sub-sequence of that object's data, as if it had been copied into a
    separate bytes-like object.  The arguments to the constructor mirror the
    arguments to the built-in :func:`slice` function, so that
    ``subbytes(data, start, stop)`` represents the same sequence of bytes as
    ``data[start:stop]``.

    .. attribute:: data

        The :attr:`data` attribute provides a reference to the underlying
        bytes-like object. If the `data` argument to the constructor is an
        instance of :class:`subbytes`, it will be unwrapped so that this
        attribute always references a bytes-like object directly.
    """

    def __init__(self,
        data: Union[bytes, "subbytes"],
        start: Optional[int] = None,
        stop: Optional[int] = None,
    ) -> None:
        self.data: bytes
        if isinstance(data, subbytes):
            self.data = data.data
            base = data
        else:
            self.data = data
            self.start = 0
            self.stop = len(data)
            base = self

        newstart = base.start if start is None else base.translate(start, True)
        newstop  = base.stop  if stop  is None else base.translate(stop,  True)

        self.start = newstart
        self.stop  = newstop

    def __bool__(self) -> bool:
        """Indicate that the sequence is non-empty."""
        return self.stop > self.start

    def __eq__(self, other: Any) -> bool:
        """Compare two sequences of bytes for equality."""
        try:
            # TypeError if other is not Sized
            if len(self) != len(other):
                return False

            # TypeError if other is not Iterable
            for left, right in zip(self, other):
                if left != right:
                    return False
        except TypeError:
            return NotImplemented

        return True

    @overload
    def __getitem__(self, key: int) -> int:
        ...

    @overload
    def __getitem__(self, key: slice) -> bytes:
        ...

    def __getitem__(self, key: Union[int, slice]) -> Union[int, bytes]:
        """Retrieve an individual byte or copy a sub-sequence."""
        if isinstance(key, int):
            key = self.translate(key)
        elif isinstance(key, slice):
            start, stop = key.start, key.stop

            key = slice(
                self.start if start is None else self.translate(start, True),
                self.stop  if stop  is None else self.translate(stop, True),
                key.step
            )

        return self.data[key]

    def __iter__(self) -> Iterator[int]:
        """Produce an iterator for this sequence."""
        for index in range(self.start, self.stop):
            yield self.data[index]

    def __len__(self) -> int:
        """Query the length of the sequence."""
        return self.stop - self.start

    def __repr__(self) -> str:
        """Provide an :func:`eval`-able representation of this object."""
        args = [repr(self.data)]

        if self.start:
            args.append(f"start={self.start}")

        if self.stop < len(self.data):
            args.append(f"stop={self.stop}")

        return f"{typename(self)}({', '.join(args)})"

    def translate(self, index: int, clamp: bool = False) -> int:
        """Convert an index of self into an index of :attr:`data`.

        This method accepts an integral `index` of any value and translates it
        into a meaningful index of :attr:`data`. If ``-len(self)`` <= `index` <
        ``len(self)``, then the result is guaranteed to be a valid index
        (meaning it will not cause an :exc:`IndexError`). If `index` <
        ``-len(self)`` and `clamp` is ``True``, then this method will return
        the index of the start of this sequence, which is only guaranteed to be
        valid if the sequence is non-empty, and may or may not be valid if the
        sequence is empty. If ``len(self)`` <= `index` and `clamp` is ``True``,
        then this method will return the stop index, which is the index just
        beyond the end of the sequence. The stop index may or may not be valid,
        regardless of the length of the sequence, but it is guaranteed to match
        the start index if the sequence is empty. Under any other scenario
        (i.e. `index` is out of range and `clamp` is ``False``), the result is
        guaranteed to be an invalid index.

        :meta private:
        """
        if index < 0:
            index += self.stop
            if index < self.start:
                if clamp:
                    return self.start
                else:
                    return len(self.data)
            else:
                return index
        else:
            index += self.start
            if index >= self.stop:
                if clamp:
                    return self.stop
                else:
                    return len(self.data)
            else:
                return index

    def consume(self) -> int:
        """Pop a byte off the front of the sequence.

        This function returns the first byte of the current sequence, advancing
        the start index relative to :attr:`data` by one. If the sequence is
        empty, then an :exc:`IndexError` will be raised.
        """
        byte = self.dereference()
        self.start += 1
        return byte

    def dereference(self) -> int:
        """Retrieve the first byte of the sequence.

        If the sequence is empty, an :exc:`IndexError` will be raised
        """
        return self[0]

    def prune(self, length: int) -> "subbytes":
        """Cut the sequence down to length.

        This method effectively splits the current sequence at index `length`,
        truncating it to that length, and returning a new :class:`subbytes`
        object that references the portion that was cut off of the end. If
        the `length` argument exceeds the length of the sequence, then the
        current object will remain unchanged and the returned :class:`subbytes`
        object will contain an empty sequence.
        """
        removed = subbytes(self, length)
        self.stop = removed.start
        return removed

    def replace(self, replacement: bytes) -> bytes:
        """Substitute the given string in place of the current sequence

        This method produces a copy of the wrapped object, in which the
        current sequence has been replaced with the contents of the
        `replacement` argument.
        """
        return self.data[:self.start] + replacement + self.data[self.stop:]

def typename(cls: Any, qualified: bool = False) -> str:
    """Query an object to determine its type.

    :param cls:
        If ``cls`` is a class, this function will return its name.  If
        ``cls`` is an object, this function will return the name of the
        object's class.
    :param bool qualified:
        Indicate whether to return the fully-qualified name, which includes
        the module path, as well as the names of any enclosing classes (for
        inner classes).
    """
    if not isinstance(cls, type):
        cls = type(cls)

    if qualified:
        return ".".join((cls.__module__, cls.__qualname__))
    else:
        return cast(type, cls).__name__



__all__ = [
    "Engine",
    "SNMPv1", "SNMPv2c", "SNMPv3",
    "noAuthNoPriv", "authNoPriv", "authPriv",
]

from snmp.engine import Engine
from snmp.message import MessageProcessingModel
from snmp.security.levels import *

SNMPv1  = MessageProcessingModel.SNMPv1
SNMPv2c = MessageProcessingModel.SNMPv2c
SNMPv3  = MessageProcessingModel.SNMPv3



__all__ = [
    "INTEGER", "OCTET_STRING", "NULL", "OBJECT_IDENTIFIER", "SEQUENCE",
    "Asn1Encodable","Primitive", "Constructed",
    "Integer", "OctetString", "Null", "OID", "Sequence",
]

from abc import abstractmethod
import re
from snmp.ber import *
from snmp.exception import *
from snmp.typing import *
from snmp.utils import *

INTEGER             = Identifier(Class.UNIVERSAL, Structure.PRIMITIVE, 2)
OCTET_STRING        = Identifier(Class.UNIVERSAL, Structure.PRIMITIVE, 4)
NULL                = Identifier(Class.UNIVERSAL, Structure.PRIMITIVE, 5)
OBJECT_IDENTIFIER   = Identifier(Class.UNIVERSAL, Structure.PRIMITIVE, 6)
SEQUENCE            = Identifier(Class.UNIVERSAL, Structure.CONSTRUCTED, 16)

Bool = Literal[False, True]
TEncodable      = TypeVar("TEncodable",     bound="Asn1Encodable")
TPrimitive      = TypeVar("TPrimitive",     bound="Primitive")
TInteger        = TypeVar("TInteger",       bound="Integer")
TOctetString    = TypeVar("TOctetString",   bound="OctetString")
TNull           = TypeVar("TNull",          bound="Null")
TOID            = TypeVar("TOID",           bound="OID")

class Asn1Encodable:
    TYPE: ClassVar[Identifier]

    def __eq__(self, other: Any) -> bool:
        if type(self) == type(other):
            return self.equals(other)
        else:
            return NotImplemented

    @overload
    @classmethod
    def decode(
        cls: Type[TEncodable],
        data: Asn1Data,
    ) -> TEncodable:
        ...

    @overload
    @classmethod
    def decode(
        cls: Type[TEncodable],
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = True,
        **kwargs: Any,
    ) -> Union[TEncodable, Tuple[TEncodable, subbytes]]:
        ...

    @classmethod
    def decode(
        cls: Type[TEncodable],
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = True,
        **kwargs: Any,
    ) -> Union[TEncodable, Tuple[TEncodable, subbytes]]:
        _copy: Bool = copy
        if leftovers:
            encoding, tail = decode(data, cls.TYPE, True, _copy)
            return cls.deserialize(encoding, **kwargs), tail
        else:
            encoding = decode(data, cls.TYPE, False, _copy)
            return cls.deserialize(encoding, **kwargs)

    def encode(self) -> bytes:
        return encode(self.TYPE, self.serialize())

    @abstractmethod
    def equals(self: TEncodable, other: TEncodable) -> bool:
        ...

    @classmethod
    @abstractmethod
    def deserialize(cls: Type[TEncodable], data: Asn1Data) -> TEncodable:
        ...

    @abstractmethod
    def serialize(self) -> bytes:
        ...

class Primitive(Asn1Encodable):
    @abstractmethod
    def appendToOID(self, oid: TOID, implied: bool = False) -> TOID:
        ...

    @classmethod
    @abstractmethod
    def decodeFromOID(
        cls: Type[TPrimitive],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TPrimitive:
        ...

class Integer(Primitive):
    TYPE = INTEGER

    BITS:       ClassVar[int]               = 32
    BYTEORDER:  ClassVar[Literal["big"]]    = "big"
    SIGNED:     ClassVar[bool]              = True

    def __init__(self, value: int) -> None:
        self._value = value

    def __repr__(self) -> str:
        return f"{typename(self)}({self.value})"

    @property
    def value(self) -> int:
        return self._value

    def equals(self, other: "Integer") -> bool:
        return self.value == other.value

    def appendToOID(self, oid: TOID, implied: bool = False) -> TOID:
        return oid.extend(self.value)

    @classmethod
    def inRange(cls, value: int) -> bool:
        assert isinstance(cls.SIGNED, bool)
        return value.bit_length() <= cls.BITS - cls.SIGNED + (value < 0)

    @classmethod
    def decodeFromOID(
        cls: Type[TInteger],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TInteger:
        value = next(nums)

        if not cls.inRange(value):
            errmsg = f"{typename(cls)} value out of range: {value}"
            raise OID.IndexDecodeError(errmsg)

        return cls(value)

    @classmethod
    def deserialize(cls: Type[TInteger], data: Asn1Data) -> TInteger:
        value = int.from_bytes(data, cls.BYTEORDER, signed=cls.SIGNED)

        if not cls.inRange(value):
            raise ParseError(f"Encoding too large for {typename(cls)}")

        return cls(value)

    def serialize(self) -> bytes:
        assert self.inRange(self.value)

        # equivalent to (N + 8) // 8
        # the reason it's not (N + 7) is that ASN.1 always includes a sign bit
        nbytes = (self.value.bit_length() // 8) + 1
        return self.value.to_bytes(nbytes, self.BYTEORDER, signed=True)

class OctetString(Primitive):
    TYPE = OCTET_STRING

    MIN_SIZE:       ClassVar[int]               = 0
    MAX_SIZE:       ClassVar[int]               = 0xffff
    INVALID_SIZES:  ClassVar[Tuple[int, ...]]   = ()

    def __init__(self, data: Asn1Data = b"") -> None:
        self._data = data

    def __repr__(self) -> str:
        return f"{typename(self)}({self.data!r})"

    @property
    def data(self) -> Asn1Data:
        return self._data

    def equals(self, other: "OctetString") -> bool:
        return self.data == other.data

    def appendToOID(self, oid: TOID, implied: bool = False) -> TOID:
        if not implied:
            oid = oid.extend(len(self.data))

        return oid.extend(*self.data)

    @classmethod
    def decodeFromOID(
        cls: Type[TOctetString],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TOctetString:
        if implied:
            length = OID.MAXLEN
        elif cls.MIN_SIZE == cls.MAX_SIZE:
            length = cls.MAX_SIZE
        else:
            length = next(nums)

        data = bytearray()
        while len(data) < length:
            try:
                byte = next(nums)
            except StopIteration:
                if implied:
                    break
                else:
                    raise

            try:
                data.append(byte)
            except ValueError as err:
                errmsg = "Sub-identifier is too large for type \"{}\": {}"
                raise OID.IndexDecodeError(
                    errmsg.format(typename(cls), byte)
                ) from err

        return cls.interpret(bytes(data))

    @classmethod
    def deserialize(cls: Type[TOctetString], data: Asn1Data) -> TOctetString:
        if len(data) < cls.MIN_SIZE:
            msg = "Encoded {} may not be less than {} bytes long"
            raise ParseError(msg.format(typename(cls), cls.MIN_SIZE))
        elif len(data) > cls.MAX_SIZE:
            msg = "Encoded {} may not be more than {} bytes long"
            raise ParseError(msg.format(typename(cls), cls.MAX_SIZE))
        elif len(data) in cls.INVALID_SIZES:
            msg = "Encoded {} not permitted to be {} bytes long"
            raise ParseError(msg.format(typename(cls), len(data)))

        return cls.interpret(data)

    @classmethod
    def interpret(cls: Type[TOctetString], data: Asn1Data) -> TOctetString:
        return cls(data)

    def serialize(self) -> bytes:
        data = self.data
        if len(data) < self.MIN_SIZE:
            msg = "Encoded {} may not be less than {} bytes long"
            raise ValueError(msg.format(typename(self), self.MIN_SIZE))
        elif len(data) > self.MAX_SIZE:
            msg = "Encoded {} may not be more than {} bytes long"
            raise ValueError(msg.format(typename(self), self.MAX_SIZE))
        elif len(data) in self.INVALID_SIZES:
            msg = "Encoded {} not permitted to be {} bytes long"
            raise ValueError(msg.format(typename(self), len(data)))

        return data[:] if isinstance(data, subbytes) else data

class Null(Primitive):
    TYPE = NULL

    def __repr__(self) -> str:
        return f"{typename(self)}()"

    def equals(self, other: "Null") -> bool:
        return True

    def appendToOID(self, oid: TOID, implied: bool = False) -> TOID:
        return oid

    @classmethod
    def decodeFromOID(
        cls: Type[TNull],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TNull:
        return cls()

    @classmethod
    def deserialize(cls: Type[TNull], data: Asn1Data) -> TNull:
        return cls()

    def serialize(self) -> bytes:
        return b""

class OID(Primitive):
    TYPE = OBJECT_IDENTIFIER

    DOT:    ClassVar[str] = "."
    MULT:   ClassVar[int] = 40
    MAXLEN: ClassVar[int] = 128

    class BadPrefix(IncomingMessageError):
        pass

    class IndexDecodeError(IncomingMessageError):
        pass

    def __init__(self, *nums: int) -> None:
        if len(nums) > self.MAXLEN:
            errmsg = "{} may not contain more than {} sub-identifiers"
            raise ValueError(errmsg.format(typename(self), self.MAXLEN))

        assert all(0 <= n < (1 << 32) for n in nums)
        self.nums = nums

    def __repr__(self) -> str:
        return f"{typename(self)}{self.nums}"

    def __str__(self) -> str:
        return self.DOT.join(str(n) for n in self.nums)

    @overload
    def __getitem__(self, idx: int) -> int:
        ...

    @overload
    def __getitem__(self, idx: slice) -> Tuple[int, ...]:
        ...

    def __getitem__(self,
        idx: Union[int, slice],
    ) -> Union[int, Tuple[int, ...]]:
        return self.nums.__getitem__(idx)

    def __hash__(self) -> int:
        return self.nums.__hash__()

    def __iter__(self) -> Iterator[int]:
        return self.nums.__iter__()

    def __len__(self) -> int:
        return self.nums.__len__()

    def __lt__(self, other: "OID") -> bool:
        return self.nums < other.nums

    @staticmethod
    def serializeSubIdentifier(bytearr: bytearray, num: int) -> None:
        if num < 0x80:
            bytearr.append(num)
        else:
            flag = 0
            tmp = bytearray()

            while num:
                tmp.append((num & 0x7f) | flag)
                flag = 0x80
                num >>= 7

            tmp.reverse()
            bytearr.extend(tmp)

    def tryDecode(self,
        nums: Iterator[int],
        cls: Type[TPrimitive],
        implied: bool = False,
    ) -> TPrimitive:
        try:
            return cls.decodeFromOID(nums, implied=implied)
        except StopIteration as err:
            errmsg = f"Incomplete {typename(cls)} index"
            raise OID.IndexDecodeError(errmsg) from err

    FIRST = re.compile(r"^\.?(\d+|$)")
    REGEX = re.compile(r"\.(\d+)")

    @classmethod
    def parse(cls: Type[TOID], oid: str) -> TOID:
        match = cls.FIRST.match(oid)

        if match is None:
            raise ValueError(f"Invalid OID string: {oid}")

        nums = []
        if match.group(1):
            while match is not None:
                nums.append(int(match.group(1)))
                index = match.end()
                match = cls.REGEX.match(oid, index)

            if index != len(oid):
                raise ValueError(f"Trailing characters in OID string: {oid}")

        try:
            if nums[0] > 2:
                errmsg = "{} may not begin with {}"
                raise ValueError(errmsg.format(typename(cls), nums[0]))

            if nums[1] >= cls.MULT:
                errmsg = "second number in {} must be less than {}"
                raise ValueError(errmsg.format(typename(cls), nums[1]))
        except IndexError as err:
            pass

        if any(n < 0 for n in nums):
            raise ValueError("\"{}\" contains a negative sub-identifier")
        elif any(n >= (1 << 32) for n in nums):
            errmsg = "OID \"{}\" contains a sub-identifier that is too large"
            raise ValueError(errmsg.format(oid))

        return cls(*nums)

    def appendIndex(self: TOID,
        *index: Primitive,
        implied: bool = False,
    ) -> TOID:
        oid = self
        for obj in index[:-1]:
            oid = obj.appendToOID(oid)

        for obj in index[-1:]:
            oid = obj.appendToOID(oid, implied=implied)

        return oid

    def extend(self: TOID, *nums: int) -> TOID:
        nums = self.nums + nums
        return type(self)(*nums)

    # TODO: I'm not sure whether the return type annotation is correct, or
    #       if the correct annotation even exists
    def extractIndex(self,
        prefix: "OID",
        *types: Type[TPrimitive],
        implied: bool = False,
    ) -> Tuple[TPrimitive, ...]:
        if len(self.nums) < len(prefix):
            errmsg = "\"{}\" is shorter than the given prefix \"{}\""
            raise self.BadPrefix(errmsg.format(self, prefix))

        if self.nums[:len(prefix)] != prefix.nums:
            errmsg = "\"{}\" does not begin with \"{}\""
            raise self.BadPrefix(errmsg.format(self, prefix))

        nums = iter(self.nums[len(prefix):])

        index = []
        for cls in types[:-1]:
            index.append(self.tryDecode(nums, cls))

        for cls in types[-1:]:
            index.append(self.tryDecode(nums, cls, implied=implied))

        try:
            next(nums)
        except StopIteration:
            pass
        else:
            errmsg = "Not all sub-identifiers were consumed"
            raise self.IndexDecodeError(errmsg)

        return tuple(index)

    def getIndex(self,
        prefix: "OID",
        cls: Type[TPrimitive] = Integer,    # type: ignore[assignment]
        implied: bool = False,
    ) -> TPrimitive:
        return self.extractIndex(prefix, cls, implied=implied)[0]

    def startswith(self, prefix: "OID") -> bool:
        return prefix.nums == self.nums[:len(prefix)]

    def equals(self, other: "OID") -> bool:
        return self.nums == other.nums

    def appendToOID(self, oid: TOID, implied: bool = False) -> TOID:
        if not implied:
            oid = oid.extend(len(self.nums))

        return oid.extend(*self.nums)

    @classmethod
    def decodeFromOID(
        cls: Type[TOID],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TOID:
        if implied:
            length = OID.MAXLEN
        else:
            length = next(nums)

        subids = []
        for i in range(length):
            try:
                subids.append(next(nums))
            except StopIteration:
                if implied:
                    break
                else:
                    raise

        return cls(*subids)

    @classmethod
    def deserialize(cls: Type[TOID], data: Asn1Data) -> TOID:
        stream = iter(data)

        try:
            oid = list(divmod(next(stream), cls.MULT))
        except StopIteration as err:
            raise ParseError(f"Empty {typename(cls)}") from err

        value = 0
        for byte in stream:
            value |= byte & 0x7f
            if byte & 0x80:
                value <<= 7
                if value >= (1 << 32):
                    raise ParseError("Sub-identifier out of range")
            else:
                oid.append(value)
                value = 0

        if value:
            raise ParseError("OID ended unexpectedly")

        return cls(*oid)

    def serialize(self) -> bytes:
        try:
            first = self.nums[0]
        except IndexError:
            return b"\x00"

        try:
            second = self.nums[1]
        except IndexError:
            second = 0

        encoding = bytearray()
        self.serializeSubIdentifier(encoding, first * self.MULT | second)
        for number in self.nums[2:]:
            self.serializeSubIdentifier(encoding, number)

        return bytes(encoding)

class Constructed(Asn1Encodable):
    def equals(self, other: "Constructed") -> bool:
        if len(self) == len(other):
            for left, right in zip(self, other):
                if left != right:
                    return False

        return True

    @abstractmethod
    def __iter__(self) -> Iterator[Asn1Encodable]:
        ...

    @abstractmethod
    def __len__(self) -> int:
        ...

    def serialize(self) -> bytes:
        return b"".join([item.encode() for item in self])

class Sequence(Constructed):
    TYPE = SEQUENCE
